#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to experiment procedures

# (c) Issabella (tutorial) / Dinasset (filter) / 2021
#
#   History:
#
#   v1.0: 2021-06-30: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import sys, os
import random
import time
import datetime
from gimpfu import *
import pygtk
import gtk
from sys import argv, platform
from ctypes import *

# define here some constant value, when required
FixWH=3000

# predefined colour names
NEUTRAL_GRAY=(170,170,170,255)
RED_GRAY=(200,170,170,255)
ORANGE_GRAY=(200,185,170,255)
YELLOW_GRAY=(200,200,170,255)
CHARTREUSE_GRAY=(185,200,170,255)
GREEN_GRAY=(170,200,170,255)
SPRING_GRAY=(170,200,185,255)
CYAN_GRAY=(170,200,200,255)
AZURE_GRAY=(170,185,200,255)
BLUE_GRAY=(170,170,200,255)
VIOLET_GRAY=(185,170,200,255)
MAGENTA_GRAY=(200,170,200,255)
ROSE_GRAY=(200,170,185,255)

############################################################################
# Service routines
#---------------------------------------------------------------------------

def load_library (library_name):
    if platform == "linux" or platform == "linux2":
        library_name = library_name + '.so.0'
    elif platform == "win32":
        from ctypes.util import find_library
        library_name = find_library (library_name + "-0")
    else:
        raise BaseException ("TODO")
    return CDLL (library_name)
#import sys
#print (sys.version)
gimp = load_library ('libgimp-2.0')
#print (gimp)
#gimp = load_library ('libgimp-3.0')

def gegl_graph(image, drawable, gegl_graph_string):
    class GeglBuffer(Structure):
        pass
    drawable_id = drawable.ID
    
    gegl = load_library ('libgegl-0.4')
    sucess = gegl.gegl_init (None, None)
    gimp.gimp_drawable_get_shadow_buffer.restype = POINTER (GeglBuffer)
    gimp.gimp_drawable_get_buffer.restype        = POINTER (GeglBuffer)

    x = c_int (); y = c_int (); w = c_int (); h = c_int ()
    non_empty,x,y,w,h = pdb.gimp_drawable_mask_intersect (drawable)
    args = [b"string", c_char_p (gegl_graph_string), c_void_p ()]
    
    if non_empty:
        source = gimp.gimp_drawable_get_buffer (drawable_id)
        target = gimp.gimp_drawable_get_shadow_buffer (drawable_id)
        sucess = gegl.gegl_render_op (source, target, "gegl:gegl", *args)
        gegl.gegl_buffer_flush (target)
        gimp.gimp_drawable_merge_shadow (drawable_id, PushUndo = True)
        gimp.gimp_drawable_update (drawable_id, x, y, w, h)
        gimp.gimp_displays_flush ()
        
# here the single GEGL functions

def gegl_c2g(image, drawable, uradius, usamples, uiter):
    radius=int(uradius)      #dft 300   #ui_range (2, 1000)
    samples=int(usamples)    #dft 4     #ui_range (3, 17)
    iterations=int(uiter)    #dft 10    #ui_range (1, 30)
    enhance_shadows=False
    mystring="gegl:c2g radius=%d samples=%d iterations=%d enhance-shadows=%d"
    gegl_graph_string=mystring % (radius, samples, iterations, enhance_shadows)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_sobel(image, drawable):   
    horizontal=True   
    vertical=True   
    keep_sign=True    
    gegl_graph_string="gegl:edge-sobel horizontal=%d vertical=%d keep-sign=%d" % (horizontal, vertical, keep_sign)
    gegl_graph(image, drawable,gegl_graph_string)

def gegl_edge(image, drawable): 
    algorithm=0     # 0=sobel,1=prewitt,2=gradient,3=roberts,4=differential,5=laplace
    amount=2.0      # value_range  (1.00 to 10.0)
    border=1        # 0=none,1=clamp,2=loop,3=black,4=white
    mystring="gegl:edge algorithm=%d amount=%f border=%d"
    gegl_graph_string=mystring % (algorithm, amount, border)
    gegl_graph(image, drawable,gegl_graph_string)
 
    
############################################################################
# include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    pdb.gimp_message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    pdb.gimp_message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
    #pdb.gimp_selection_none (inImage)
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    optiWidthHeight = FixWH       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)
    pdb.gimp_selection_none (newImage)
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "NewImage: Background Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # adjust mean values
    adjustmeanv (newImage, newBG)
    # remove low values, remove high values  
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                   1.00, 0.10, 0.90, False)
    newBG.opacity = 100.0
    
    return(newImage)

#---------------------------------------------------------------------------
# adjust values
#---------------------------------------------------------------------------
def adjustmeanv(inImage, inLayer):
    # new version modified to take care of the existing mean in determining the theretical mean 
    # check mean value of the layer
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
    # consider 3 classes
    if meanl < 85: thmean = 100
    elif meanl > 170: thmean = 155
    else: thmean = 127.5    # regular
    minmean = thmean-16
    maxmean = thmean+16
    
    modified=False
    if meanl < minmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl)+" ratio = "+str(ratio))
        while meanl < minmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        modified=True
        #info_msg("increased mean value = "+str(meanl))
    if meanl > maxmean:
        ratio = thmean/meanl*1.0 
        #info_msg("mean value found = "+str(meanl))
        while meanl > maxmean:  
            pdb.gimp_drawable_curves_spline (inLayer, HISTOGRAM_VALUE, 8,
                                             (0.0,0.0,0.3,0.3*ratio, 0.7,0.7*ratio, 1.0,1.0))
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (inLayer,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("reduced mean value = "+str(meanl))
    
    return(inLayer)
     
# ------------------------------------------------------
# convert RGB to HSV 
# ------------------------------------------------------

def rgb2hsv(rgb):
    (r,g,b,a) = rgb
    r, g, b = r/255.0, g/255.0, b/255.0
    mx = max(r, g, b)
    mn = min(r, g, b)
    df = mx-mn
    if mx == mn:
        h = 0
    elif mx == r:
        h = (60 * ((g-b)/df) + 360) % 360
    elif mx == g:
        h = (60 * ((b-r)/df) + 120) % 360
    elif mx == b:
        h = (60 * ((r-g)/df) + 240) % 360
    if mx == 0:
        s = 0
    else:
        s = (df/mx)*100
    v = mx*100
    #info_msg ("hue="+str(h))
    #info_msg ("sat="+str(s))
    #info_msg ("val="+str(v))
    return h, s, v

# ------------------------------------------------------
# create paper for the BG 
# ------------------------------------------------------
def CreatePaper (width,height,hue,sat,val):
    
    WH = width+height
    PaperImg = pdb.gimp_image_new (width, height, RGB_IMAGE)
    PaperDrw = pdb.gimp_layer_new (PaperImg, width, height, RGB_IMAGE,
                                   "Generated Paper", 100.0, LAYER_MODE_NORMAL)
    PaperImg.add_layer (PaperDrw,0)
    
    pdb.plug_in_plasma (PaperImg, PaperDrw, 26, 2.0)
    pdb.gimp_drawable_curves_spline (PaperDrw, HISTOGRAM_VALUE, 4,
                                    (0.00,0.75, 1.00,1.00) 
                                    )
    pdb.plug_in_rgb_noise (PaperImg, PaperDrw, True, False,
                           0.2, 0.2, 0.2, 0.2)
    pdb.plug_in_gauss (PaperImg, PaperDrw, 5.0, 5.0, 0)
    pdb.gimp_drawable_desaturate (PaperDrw, DESATURATE_LUMA)
    pdb.gimp_drawable_colorize_hsl (PaperDrw, hue*1.0, sat*1.0, val*1.0-100.0)
    
    #pdb.gimp_display_new (PaperImg)
    
    return (PaperDrw)
#
# ================ *
#   M  A  I  N     *
# ---------------- *
   

############################################################################
# Main
#---------------------------------------------------------------------------

def JvidEffectB (inImage, inLayer, widthheight, threshold,
                 predefcol, inColor, framepat,
                 doFlatten):
    
    # house-keeping
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # display messages using error console
    
    newImage = createOutimage (inImage, inLayer)
    Original = newImage.layers[1]
    Original.name = "Original"
    Preprocessed = newImage.layers[0]
    Preprocessed.name = "Preprocessed"
    width = Preprocessed.width
    height = Preprocessed.height
    
    #pdb.gimp_display_new (newImage)
    #return
    
    C2G = Preprocessed.copy()
    newImage.add_layer (C2G, 0)
    C2G.name = "C2G"
    # scaling down and back to improve c2g performance
    pdb.plug_in_unsharp_mask (newImage, C2G, 10.0, 2.0, 0)
    pdb.gimp_layer_scale (C2G, width/2, height/2, True)         # scale down
    gegl_c2g(newImage, C2G, 1000, 15, 10)         # good but slow
    pdb.gimp_layer_scale (C2G, width, height, True)             # scale back
    BLACK=1
    pdb.plug_in_erode (newImage, C2G, BLACK, 7, 1.0, 15, 0, 255)
    
    #pdb.gimp_display_new (newImage)
    #return
     
    Sobel = C2G.copy()
    newImage.add_layer (Sobel, 0)
    Sobel.name = "Sobel"
   #gegl_sobel (newImage, Sobel)    # repl by below 25/05
    gegl_edge (newImage, Sobel)
    
    pdb.gimp_drawable_invert (Sobel, True)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    pdb.gimp_drawable_threshold (Sobel, HISTOGRAM_VALUE, threshold*1.0/255.0, 1.0)
    
    #pdb.gimp_display_new (newImage)
    #return
    
    pdb.plug_in_colortoalpha (newImage, Sobel, (1.0,1.0,1.0,1.0))
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # ================================================================
    #  create a sheet of paper
    # ----------------------------------------------------------------
    # colour selection
    if predefcol == 0:
        (uR,uG,uB,A)=inColor
        pdb.gimp_context_set_foreground (inColor);
        po=" paper specified by user ("+str(int(uR))+","+str(int(uG))+","+str(int(uB))+")"
    if predefcol == 1:
        pdb.gimp_context_set_foreground (NEUTRAL_GRAY); po=" paper NEUTRAL_GRAY"
    if predefcol == 2:
        pdb.gimp_context_set_foreground (RED_GRAY); po=" paper RED_GRAY"
    if predefcol == 3:
        pdb.gimp_context_set_foreground (ORANGE_GRAY); po=" paper ORANGE_GRAY"
    if predefcol == 4:
        pdb.gimp_context_set_foreground (YELLOW_GRAY); po=" paper YELLOW_GRAY"
    if predefcol == 5:
        pdb.gimp_context_set_foreground (CHARTREUSE_GRAY); po=" paper CHARTREUSE_GRAY"
    if predefcol == 6:
        pdb.gimp_context_set_foreground (GREEN_GRAY); po=" paper GREEN_GRAY"
    if predefcol == 7:
        pdb.gimp_context_set_foreground (SPRING_GRAY); po=" paper SPRING_GRAY"
    if predefcol == 8:
        pdb.gimp_context_set_foreground (CYAN_GRAY); po=" paper CYAN_GRAY"
    if predefcol == 9:
        pdb.gimp_context_set_foreground (AZURE_GRAY); po=" paper AZURE_GRAY"
    if predefcol == 10:
        pdb.gimp_context_set_foreground (BLUE_GRAY); po=" paper BLUE_GRAY"
    if predefcol == 11:
        pdb.gimp_context_set_foreground (VIOLET_GRAY); po=" paper VIOLET_GRAY"
    if predefcol == 12:
        pdb.gimp_context_set_foreground (MAGENTA_GRAY); po=" paper MAGENTA_GRAY"
    if predefcol == 13:
        pdb.gimp_context_set_foreground (ROSE_GRAY); po=" paper ROSE_GRAY"
    # obtain HSV from Colour RGB to colorize paper
    (R,G,B,A)=pdb.gimp_context_get_foreground()
    hue,sat,val = rgb2hsv ((R,G,B,A))
    # create the paper with the user input colour
    Paper = CreatePaper (width, height, hue, sat, val)
    PaperBG = pdb.gimp_layer_new_from_drawable (Paper, newImage)
    newImage.add_layer (PaperBG, 1)
    PaperBG.name = "Generated Paper"
    PaperBG.mode = LAYER_MODE_NORMAL
    PaperBG.opacity = 100.0
    
    #pdb.gimp_display_new (newImage)
    #return
    
    Visible = pdb.gimp_layer_new_from_visible (newImage, newImage,
                                               "Line Art on Paper")
    newImage.add_layer (Visible, 0)
    
    (PW,PH,Pbpp) = pdb.gimp_pattern_get_info (framepat)
    Frame = pdb.gimp_layer_new (newImage, PW, PH, RGBA_IMAGE,
                                "Frame", 100.0, LAYER_MODE_NORMAL)
    newImage.add_layer (Frame, 0)
    pdb.gimp_context_set_pattern (framepat)
    pdb.gimp_edit_fill (Frame, PATTERN_FILL)
    pdb.gimp_layer_scale (Frame, width, height, False)
     
    pdb.gimp_context_set_sample_threshold_int (127)
    pdb.gimp_image_select_color (newImage, CHANNEL_OP_REPLACE,
                                 Frame, (0.0,0.0,0.0,1.0))
    pdb.gimp_edit_cut (Frame)
    pdb.gimp_selection_invert (newImage)
    pdb.gimp_edit_fill (Frame, FILL_FOREGROUND)
    
    pdb.gimp_selection_none (newImage)
    
    Frame.mode = LAYER_MODE_DISSOLVE
    Frame.opacity = 50.0
    
    pdb.gimp_display_new (newImage)
    return
    
    
    # final step
    
    if doFlatten==True:
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = "outcome" 
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+" "+oname+" with JVID effect B () " 
    
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="issabella/dinasset"
menu="<Image>/Jvid/"
copyrightYear="2021"
desc="JVID Effect B"

register(
    "JvidEffectB",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+" ...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+"\n (0=same as input)",
                                    3000, (0, 6000, 100)),
        (PF_ADJUSTMENT, "threshold", "Threshold for B&W Edges (2 to 191)",
                                    2, (2, 191, 1)),
        (PF_OPTION, "precol", "Predefined paper colour", 4, ["*user (choose below)",
                                                        "NEUTRAL_GRAY=(170,170,170,255)",
                                                        "RED_GRAY=(200,170,170,255)",
                                                        "ORANGE_GRAY=(200,185,170,255)",
                                                        "YELLOW_GRAY=(200,200,170,255)",
                                                        "CHARTREUSE_GRAY=(185,200,170,255)",
                                                        "GREEN_GRAY=(170,200,170,255)",
                                                        "SPRING_GRAY=(170,200,185,255)",
                                                        "CYAN_GRAY=(170,200,200,255)",
                                                        "AZURE_GRAY=(170,185,200,255)",
                                                        "BLUE_GRAY=(170,170,200,255)",
                                                        "VIOLET_GRAY=(185,170,200,255)",
                                                        "MAGENTA_GRAY=(200,170,200,255)",
                                                        "ROSE_GRAY=(200,170,185,255)"
                                                        ]),
        (PF_COLOR, "pcolor", "User paper color"+
                           "\n (use light colours)", (200,200,200,255)),
        (PF_PATTERN, "frame", "User paper frame", "JVID_B_PictureFadeBorder.jpg"),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    JvidEffectB,
    menu=menu
)

main()
